FREE = 0
OPEN = 1
CLOSED = 2
CANCELED = 3
ASSIGNED = 4
FAILED = 5


# Groups:
ACTIVE_GROUP = (FREE, OPEN, ASSIGNED)
DONE_GROUP = (CLOSED, CANCELED, FAILED)


def to_str(number):
    """
    Convert a task state ID number to a string.

    :param int number: task state ID, eg. 1
    :returns: state name like eg. "OPEN", or "(unknown)" if we don't know the
              name of this task state ID number.
    """
    states = globals()
    for name, value in states.items():
        if number == value and name.isalpha() and name.isupper():
            return name
    return '(unknown state %d)' % number



import errno
import os
from twisted.internet import defer


class Cache(object):
    def __init__(self, connection, directory=None):
        """
        Read-through cache manager for user and tag names.

        This cache will write everything into XDG_CACHE_HOME, or ~/.cache.

        This class does no eviction or expiration - all the data will persist
        in the cache forever. If you want to clear the cache, you'll need to
        delete files manually for now.

        :param connection: txkoji.Connection
        :param directory: optional, directory on disk to store cache data.
        """
        self.connection = connection
        self.directory = directory
        if self.directory is None:
            xdg_cache_home = os.getenv('XDG_CACHE_HOME')
            if xdg_cache_home:
                self.directory = os.path.join(xdg_cache_home, 'txkoji')
            else:
                self.directory = os.path.expanduser('~/.cache/txkoji')

    def get_name(self, type_, id_):
        """
        Read a cached name if available.

        :param type_: str, "owner" or "tag"
        :param id_: int, eg. 123456
        :returns: str, or None
        """
        cachefile = self.filename(type_, id_)
        try:
            with open(cachefile, 'r') as f:
                return f.read()
        except (OSError, IOError) as e:
            if e.errno != errno.ENOENT:
                raise

    def put_name(self, type_, id_, name):
        """
        Write a cached name to disk.

        :param type_: str, "user" or "tag"
        :param id_: int, eg. 123456
        :returns: None
        """
        cachefile = self.filename(type_, id_)
        dirname = os.path.dirname(cachefile)
        try:
            os.makedirs(dirname)
        except OSError as e:
            if e.errno != errno.EEXIST:
                raise
        with open(cachefile, 'w') as f:
            f.write(name)

    def filename(self, type_, id_):
        """
        cache filename to read for this type/id.

        :param type_: str, "user" or "tag"
        :param id_: int, eg. 123456
        :returns: str
        """
        profile = self.connection.profile
        return os.path.join(self.directory, profile, type_, str(id_))

    @defer.inlineCallbacks
    def get_or_load_name(self, type_, id_, method):
        """
        read-through cache for a type of object's name.

        If we don't have a cached name for this type/id, then we will query the
        live Koji server and store the value before returning.

        :param type_: str, "user" or "tag"
        :param id_: int, eg. 123456
        :param method: function to call if this value is not in the cache.
                       This method must return a deferred that fires with an
                       object with a ".name" attribute.
        :returns: deferred that when fired returns a str, or None
        """
        name = self.get_name(type_, id_)
        if name is not None:
            defer.returnValue(name)
        instance = yield method(id_)
        if instance is None:
            defer.returnValue(None)
        self.put_name(type_, id_, instance.name)
        defer.returnValue(instance.name)

    def user_name(self, id_):
        """
        read-through cache for a user name.

        :param id_: int, eg. 123456
        :returns: deferred that when fired returns a str, or None
        """
        return self.get_or_load_name('user', id_, self.connection.getUser)

    def tag_name(self, id_):
        """
        read-through cache for a tag name.

        :param id_: int, eg. 123456
        :returns: deferred that when fired returns a str, or None
        """
        return self.get_or_load_name('tag', id_, self.connection.getTag)



from munch import Munch
from twisted.internet import defer


class Channel(Munch):

    def hosts(self, **kwargs):
        """
        Convenience wrapper around listHosts(...) for this channel ID.

        :param **kwargs: keyword arguments to the listHosts RPC.
        :returns: deferred that when fired returns a list of hosts (dicts).
        """
        kwargs['channelID'] = self.id
        return self.connection.listHosts(**kwargs)

    def tasks(self, **opts):
        """
        Convenience wrapper around listTasks(...) for this channel ID.

        Tasks are sorted by priority and creation time.

        :param **opts: "opts" dict to the listTasks RPC.
                       For example, "state=[task_states.OPEN]" will return
                       only the "OPEN" tasks.
        :returns: deferred that when fired returns a list of Tasks.
        """
        opts['channel_id'] = self.id
        qopts = {'order': 'priority,create_time'}
        return self.connection.listTasks(opts, qopts)

    @defer.inlineCallbacks
    def total_capacity(self):
        """
        Find the total task capacity available for this channel.

        Query all the enabled hosts for this channel and sum up all the
        capacities.

        Each task has a "weight". Each task will be in "FREE" state until
        there is enough capacity for the task's "weight" on a host.

        :returns: deferred that when fired returns a float value: the total
                  task weight that this channel can have open simultaneously.
        """
        # Ensure this task's channel has spare capacity for this task.
        total_capacity = 0
        hosts = yield self.hosts(enabled=True)
        for host in hosts:
            total_capacity += host.capacity
        defer.returnValue(total_capacity)



from datetime import timedelta
from txkoji import build_states
from twisted.internet import defer


"""
Methods for estimating build times.
"""


def average_build_duration(connection, package):
    """
    Return the average build duration for a package (or container).

    :param connection: txkoji.Connection
    :param package: package name
    :returns: deferred that when fired returns a datetime.timedelta object
    """
    if package.endswith('-container'):
        return average_last_builds(connection, package)
    return connection.getAverageBuildDuration(package)


def average_build_durations(connection, packages):
    """
    Return the average build duration for a package (or container).

    :param connection: txkoji.Connection
    :param list packages: package names
    :returns: deferred that when fired returns a list of timdelta objects
    """
    deferreds = [average_build_duration(connection, pkg) for pkg in packages]
    return defer.gatherResults(deferreds, consumeErrors=True)


@defer.inlineCallbacks
def average_last_builds(connection, package, limit=5):
    """
    Find the average duration time for the last couple of builds.

    :param connection: txkoji.Connection
    :param package: package name
    :returns: deferred that when fired returns a datetime.timedelta object, or
              None if there were no previous builds for this package.
    """
    # TODO: take branches (targets, or tags, etc) into account when estimating
    # a package's build time.
    state = build_states.COMPLETE
    opts = {'limit': limit, 'order': '-completion_time'}
    builds = yield connection.listBuilds(package, state=state, queryOpts=opts)
    if not builds:
        defer.returnValue(None)
    durations = [build.duration for build in builds]
    average = sum(durations, timedelta()) / len(durations)
    # print('average duration for %s is %s' % (package, average))
    defer.returnValue(average)



from twisted.internet.ssl import optionsForClientTLS
from twisted.internet._sslverify import IOpenSSLTrustRoot
from twisted.web.client import BrowserLikePolicyForHTTPS
from zope.interface import implementer

"""
This module provides some helper classes for working with Client SSL
authentication and Twisted.

There is no simple way to trust an SSL certificate bundle with Twisted. Twisted
exposes "trustRoot" that expects a trustRootFromCertificates (list of
Certificate objects), but Certificate.loadPEM() can only load a single PEM file
at a time. This means we cannot do something simple like
Certificate.loadPEM('/etc/pki/tls/certs/ca-bundle.trust.crt').

https://stackoverflow.com/questions/26166444/twisted-python-how-to-create-a-twisted-web-client-browserlikepolicyforhttps-with

The solution is to implement our own RootCATrustRoot trustRoot class that
points at a single SSL root bundle.

Warning: this uses several private Twisted interfaces, (_sslverify and
IOpenSSLTrustRoot), so this might break with future Twisted releases.

If this ends up being too fragile, we could just take this out and require
users to trust their custom Koji CAs system-wide. Alternatively, OpenSSL will
load a custom CA from the SSL_CERT_FILE environmnet variable.
"""


class ClientCertPolicy(BrowserLikePolicyForHTTPS):
    """ SSL client policy that does client authentication. """
    def __init__(self, trustRoot, client_cert):
        self.client_cert = client_cert
        super(ClientCertPolicy, self).__init__(trustRoot)

    def creatorForNetloc(self, hostname, port):
        return optionsForClientTLS(hostname=hostname.decode("ascii"),
                                   clientCertificate=self.client_cert)


@implementer(IOpenSSLTrustRoot)
class RootCATrustRoot(object):
    """
    This class trusts one particular CA bundle file in the SSL context.

    (See OpenSSLDefaultPaths for an alternate example, where it loads all the
    OpenSSL default verify paths instead.)

    Pass an instance of this class as the trustRoot argument to
    BrowserLikePolicyForHTTPS (or ClientCertPolicy).
    """
    def __init__(self, root_certificate_path):
        self.root_certificate_path = root_certificate_path

    def _addCACertsToContext(self, context):
        context.load_verify_locations(self.root_certificate_path)



class Call(object):
    """
    Callable abstract class representing a Koji RPC, eg "getTag".

    :param connection: ``txkoji.Connection``
    :param name: XML-RPC method name to call on the server, eg. "getTag".
    """
    def __init__(self, connection, name):
        self.connection = connection
        self.name = name

    def __call__(self, *args, **kwargs):
        return self.connection.call(self.name, *args, **kwargs)



import json
from txkoji.build import Build
from txkoji.task import Task
from twisted.internet import defer
from twisted.python.compat import StringType

"""
A set of "event" classes representing the messages types that Koji publishes.

Features:

* Standardize ingestion of Stompest Frames / JSON bodies with the
  "from_frame()" method. Eventually we could support other Python messaging
  libraries here.

* Normalize the "owner/owner_id/owner_name" stuff to a single "user()"
  method.

* Unified "tag()" method for all TaskStateChange and TagUntag events.
"""

# See Koji's plugins/hub/messagebus.py for the events that Koji announces.


class BuildStateChange(object):
    def __init__(self, build, event):
        self.build = build
        self.event = event  # str, eg "COMPLETE". See txkoji.build_states.

    @classmethod
    def from_frame(klass, frame, connection):
        """
        Create a new BuildStateChange event from a Stompest Frame.
        """
        event = frame.headers['new']
        data = json.loads(frame.body)
        info = data['info']
        build = Build.fromDict(info)
        build.connection = connection
        return klass(build, event)

    @property
    def url(self):
        """ Return a kojiweb URL for this change. """
        return self.build.url

    def user(self):
        """ Return a (deferred) Koji user name for this change. """
        # All BuildStateChange messages *should* have owner_name populated now.
        # See BREW-1640.
        return defer.succeed(self.build.owner_name)


class TaskStateChange(object):
    def __init__(self, task, event):
        self.task = task
        self.event = event  # str, eg "FREE". See txkoji.task_states.

    @classmethod
    def from_frame(klass, frame, connection):
        """
        Create a new TaskStateChange event from a Stompest Frame.
        """
        event = frame.headers['new']
        data = json.loads(frame.body)
        info = data['info']
        task = Task.fromDict(info)
        task.connection = connection
        return klass(task, event)

    def tag(self):
        """ Return a (deferred) cached Koji tag name for this change. """
        name_or_id = self.task.tag
        if name_or_id is None:
            return defer.succeed(None)
        if isinstance(name_or_id, StringType):
            return defer.succeed(name_or_id)
        if isinstance(name_or_id, int):
            return self.task.connection.cache.tag_name(name_or_id)
        return defer.fail()

    @property
    def url(self):
        """ Return a kojiweb URL for this change. """
        return self.task.url

    def user(self):
        """ Return a (deferred) cached Koji user name for this change. """
        # Note, do any tasks really have an "owner_id", or are they all
        # "owner"?
        owner_id = getattr(self.task, 'owner_id', self.task.owner)
        return self.task.connection.cache.user_name(owner_id)


class TagUntag(object):
    """ Tagging or Untagging an existing build. """
    def __init__(self, build, event, tag, user):
        self.build = build
        self.event = event  # str, eg "Tag" or "Untag"
        self.tag = tag  # str, tag name
        self._user = user  # str, user name

    @classmethod
    def from_frame(klass, frame, connection):
        """
        Create a new TagUntag event from a Stompest Frame.
        """
        event = frame.headers['type']  # "Tag" / "Untag"
        tag = frame.headers['tag']
        user = frame.headers['user']
        data = json.loads(frame.body)
        info = data['build']
        build = Build.fromDict(info)
        build.connection = connection
        return klass(build, event, tag, user)

    @property
    def url(self):
        """ Return a kojiweb URL for this change. """
        return self.build.url

    def user(self):
        """ Return a (deferred) Koji user name for this change. """
        return defer.succeed(self._user)



# TODO: use https://pypi.python.org/pypi/enum34

BUILDING = 0
COMPLETE = 1
DELETED = 2
FAILED = 3
CANCELED = 4


def to_str(number):
    """
    Convert a build state ID number to a string.

    :param int number: build state ID, eg. 1
    :returns: state name like eg. "BUILDING", or "(unknown)" if we don't know
              the name of this build state ID number.
    """
    states = globals()
    for name, value in states.items():
        if number == value and name.isalpha() and name.isupper():
            return name
    return '(unknown state %d)' % number



from datetime import datetime, timedelta
import os.path
import posixpath
from munch import Munch, unmunchify
from twisted.internet import defer
from txkoji import task_states
from txkoji.channel import Channel
from txkoji.estimates import average_build_duration
try:
    from urllib.parse import urlparse
    import xmlrpc
except ImportError:
    from urlparse import urlparse
    import xmlrpclib as xmlrpc


# The default kojid sleeptime upstream is 15. The RH builders have this dialed
# up to 45. We'll pick the higher number for estimation purposes.
SLEEPTIME = timedelta(seconds=45)

# A lot of the .params parsing here is conceptually similar to Koji's
# _do_parseTaskParams in the CLI.


class Task(Munch):

    @property
    def arch(self):
        """
        Return an architecture for this task.

        :returns: an arch string (eg "noarch", or "ppc64le"), or None this task
                  has no architecture associated with it.
        """
        if self.method in ('buildArch', 'createdistrepo', 'livecd'):
            return self.params[2]
        if self.method in ('createrepo', 'runroot'):
            return self.params[1]
        if self.method == 'createImage':
            return self.params[3]
        if self.method == 'indirectionimage':
            return self.params[0]['arch']

    @property
    def arches(self):
        """
        Return a list of architectures for this task.

        :returns: a list of arch strings (eg ["ppc64le", "x86_64"]). The list
                  is empty if this task has no arches associated with it.
        """
        if self.method == 'image':
            return self.params[2]
        if self.arch:
            return [self.arch]
        return []

    @property
    def build_id(self):
        """
        Find a build ID for this task.

        Pass this build ID into connection.getBuild() to get the Build class
        for this task.

        This only works for the tagBuild tasks. Alternatively, you can also
        locate the builds for other types of tasks using the listBuilds RPC
        with the taskID kwarg, like listBuilds(taskID=...)

        :returns: a build ID int, or None if no build is associated with this
                  task.
        """
        if self.method == 'tagBuild':
            return self.params[1]

    @property
    def channel(self):
        """
        Return the Channel for this task.

        :returns: an instance of the Channel class.
        """
        return Channel({'id': self.channel_id, 'connection': self.connection})

    @property
    def completed(self):
        """
        Return a parsed completion datetime for a task.

        :returns: a datetime object for the time this task finished,
                  or None if the task is not completed.
        """
        if not self.completion_ts:
            return None
        return datetime.utcfromtimestamp(self.completion_ts)

    @property
    def created(self):
        """
        Return a parsed created datetime for a task.

        :returns: a datetime object for the time this task was created
        """
        return datetime.utcfromtimestamp(self.create_ts)

    @property
    def started(self):
        """
        Return a parsed started datetime for a task.

        :returns: a datetime object for the time this task started
        """
        if not self.start_ts:
            return None
        return datetime.utcfromtimestamp(self.start_ts)

    @property
    def state_name(self):
        """
        Return a human-readable name of this task's state.

        :returns: eg. "OPEN"
        """
        return task_states.to_str(self.state)

    @property
    def duration(self):
        """
        Return a timedelta for this task.

        Measure the time between this task's start and end time, or "now"
        if the task has not yet finished.

        :returns: timedelta object, or None if the task has not even started.
        """
        if not self.started:
            return None
        start = self.started
        end = self.completed
        if not end:
            end = datetime.utcnow()
        return end - start

    @defer.inlineCallbacks
    def estimate_completion(self):
        """
        Estimate completion time for a task.

        :returns: deferred that when fired returns a datetime object for the
                  estimated, or the actual datetime, or None if we could not
                  estimate a time for this task method.
        """
        if self.completion_ts:
            # Task is already complete. Return the exact completion time:
            defer.returnValue(self.completed)
        # Get the timestamps from the descendent task that's doing the work:
        if self.method == 'build' or self.method == 'image':
            subtask_completion = yield self.estimate_descendents()
            defer.returnValue(subtask_completion)
        if self.state == task_states.FREE:
            est_completion = yield self._estimate_free()
            defer.returnValue(est_completion)
        avg_delta = yield self.estimate_duration()
        if avg_delta is None:
            defer.returnValue(None)
        est_completion = self.started + avg_delta
        defer.returnValue(est_completion)

    def estimate_duration(self):
        """
        Estimate duration (timedelta) for this task.

        Estimate the average length of time we expect between this task's
        start and end times.

        :returns: deferred that when fired returns a timedelta object for the
                  estimated timedelta, or the actual timedelta, or None if we
                  could not estimate a time for this task method.
        """
        if self.completion_ts:
            # Task is already complete. Return the exact duration timedelta.
            return defer.succeed(self.duration)
        if not self.package:
            # TODO: estimate duration some other way besides
            # getAverageBuildDuration. For example, we could estimate
            # completion time for newRepo/createrepo tasks by looking back at
            # the the past couple of tasks for this tag.
            return defer.succeed(None)
        if self.method == 'tagBuild':
            # These are pretty short. Haphazardly guess the max SLEEPTIME plus
            # a few seconds.
            tag_build_time = SLEEPTIME + timedelta(seconds=15)
            return defer.succeed(tag_build_time)
        return average_build_duration(self.connection, self.package)

    @defer.inlineCallbacks
    def _estimate_free(self):
        """
        Estimate completion time for a free task.

        :returns: deferred that when fired returns a datetime object for the
                  estimated, or the actual datetime, or None if we could not
                  estimate a time for this task method.
        """
        # Query the information we need for this task's channel and package.
        capacity_deferred = self.channel.total_capacity()
        open_tasks_deferred = self.channel.tasks(state=[task_states.OPEN])
        avg_delta_deferred = self.estimate_duration()
        deferreds = [capacity_deferred,
                     open_tasks_deferred,
                     avg_delta_deferred]
        results = yield defer.gatherResults(deferreds, consumeErrors=True)
        capacity, open_tasks, avg_delta = results
        # Ensure this task's channel has spare capacity for this task.
        open_weight = sum([task.weight for task in open_tasks])
        if open_weight >= capacity:
            # TODO: Evaluate all tasks in the channel and
            # determine when enough OPEN tasks will complete so that we can
            # get to OPEN.
            raise NotImplementedError('channel %d is at capacity' %
                                      self.channel_id)
        # A builder will pick up this task and start it within SLEEPTIME.
        # start_time is the maximum amount of time we expect to wait here.
        start_time = self.created + SLEEPTIME
        if avg_delta is None:
            defer.returnValue(None)
        est_completion = start_time + avg_delta
        defer.returnValue(est_completion)

    @defer.inlineCallbacks
    def estimate_descendents(self):
        """
        Estimate from the descendent (child) tasks.

        :returns: deferred that when fired returns a datetime object for the
                  estimated, or actual datetime, or None if there is no support
                  for this task method. Currently the only supported methods
                  here are "build" and "image".
        :raises NoDescendentsError: If we expected to find descendents for this
                  task, but there are none open. Possible explanations:
                  * Koji has not assigned this task to a worker, because it's
                    over capacity, or because it takes a few seconds to assign.
                    You may see descendant tasks in FREE state here, instead of
                    OPEN state.
                  * The makeSRPMFromSCM descendent task for this build task is
                    not yet complete.
                  * The tagBuild descendent task for this build task is not yet
                    complete.
                  If you hit this NoDescendentsError, you may want to try again
                  in a few minutes.
        """
        child_method = None
        if self.method == 'build':
            child_method = 'buildArch'
        if self.method == 'image':
            child_method = 'createImage'
        if child_method is None:
            defer.returnValue(None)
        # Find the open child task and estimate that.
        subtasks = yield self.descendents(method=child_method,
                                          state=task_states.OPEN)
        if not subtasks:
            raise NoDescendentsError('no running %s for task %d' %
                                     (child_method, self.id))
        # Find subtask with the most recent start time:
        build_task = subtasks[0]
        for subtask in subtasks:
            if subtask.start_ts > build_task.start_ts:
                build_task = subtask
        subtask_completion = yield build_task.estimate_completion()
        defer.returnValue(subtask_completion)

    @defer.inlineCallbacks
    def descendents(self, method=None, state=None):
        """
        Find descendant tasks, optionally filtered by method and/or state.

        :param method: (optional) filter for tasks, eg. "buildArch".
        :param state: (optional) filter for tasks, eg. task_states.OPEN.
        :returns: deferred that when fired returns a list of Tasks.
        """
        subtasks = yield self.connection.getTaskDescendents(self.id)
        if method:
            subtasks = [t for t in subtasks if t.method == method]
        if state:
            subtasks = [t for t in subtasks if t.state == state]
        defer.returnValue(subtasks)

    @property
    def package(self):
        """
        Find a package name from a build task's parameters.

        :returns: name of the package this build task is building.
        :raises: ValueError if we could not parse this tasks's request params.
        """
        if self.method == 'buildNotification':
            return self.params[1]['name']
        if self.method in ('createImage', 'image', 'livecd'):
            return self.params[0]
        if self.method == 'indirectionimage':
            return self.params[0]['name']
        # params[0] is the source URL for these tasks:
        if self.method not in ('build', 'buildArch', 'buildContainer',
                               'buildMaven', 'buildSRPMFromSCM', 'maven'):
            return None
        # (I wish there was a better way to do this.)
        source = self.params[0]
        o = urlparse(source)
        # build tasks can load an SRPM from a "cli-build" tmpdir:
        if source.endswith('.src.rpm'):
            srpm = os.path.basename(source)
            (name, version, release) = srpm.rsplit('-', 2)
            # Note we're throwing away version and release here. They could be
            # useful eventually, maybe in a "Package" class.
            return name
        # or an allowed SCM:
        elif o.scheme:
            package = os.path.basename(o.path)
            if package.endswith('.git'):
                package = package[:-4]
            if self.method == 'buildContainer':
                package += '-container'
            return package
        raise ValueError('could not parse source "%s"' % source)

    @property
    def params(self):
        """
        Return a list of parameters in this task's request.

        If self.request is already a list, simply return it.

        If self.request is a raw XML-RPC string, parse it and return the
        params.
        """
        if isinstance(self.request, list):
            return unmunchify(self.request)
        (params, _) = xmlrpc.loads(self.request)
        return params

    @property
    def is_scratch(self):
        for param in self.params:
            if isinstance(param, dict):
                if param.get('scratch'):
                    return True
        return False

    @property
    def tag(self):
        """
        Return the tag's name (or id number) for this task.

        :returns: An int (tag id) or string (tag name, eg "foo-build").
                  This seems to depend on the task method. For example,
                  buildArch, tagBuild, and tagNotification tasks always return
                  a tag ID here.
                  If you do get an int back here, you'll have to make a
                  separate getTag RPC to get the tag's name.
        """
        if self.method == 'buildArch':
            # Note: buildArch tag will be an int here.
            return self.params[1]
        if self.method in ('createdistrepo', 'distRepo', 'newRepo', 'runroot',
                           'tagBuild', 'waitrepo'):
            return self.params[0]
        if self.method == 'tagNotification':
            return self.params[2]
        if self.method == 'buildMaven':
            return self.params[1]['name']

    @property
    def target(self):
        if self.method in ('build', 'buildContainer', 'chainmaven', 'maven'):
            return self.params[1]
        if self.method == 'buildNotification':
            if self.params[2]:
                return self.params[2]['name']
        if self.method == 'createImage':
            return self.params[4]['name']
        if self.method in ('image', 'livecd'):
            return self.params[3]
        if self.method == 'indirectionimage':
            return self.params[0]['target']
        if self.method == 'wrapperRPM':
            return self.params[1]['name']

    @property
    def url(self):
        """
        Return a kojiweb URL for this resource.

        :returns: ``str``, kojiweb URL like
                  "http://cbs.centos.org/koji/taskinfo?taskID=381617"
        """
        endpoint = 'taskinfo?taskID=%d' % self.id
        return posixpath.join(self.connection.weburl, endpoint)


class NoDescendentsError(Exception):
    """ Could not find open buildArch descendents for this task. """
    pass
